# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/command/v1/storage_commands.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/command/v1/storage_commands.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import storage_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_storage__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.olvid/daemon/command/v1/storage_commands.proto\x12\x17olvid.daemon.command.v1\x1a\'olvid/daemon/datatypes/v1/storage.proto\"m\n\x12StorageListRequest\x12L\n\x06\x66ilter\x18\x01 \x01(\x0b\x32/.olvid.daemon.datatypes.v1.StorageElementFilterH\x00R\x06\x66ilter\x88\x01\x01\x42\t\n\x07_filter\"\\\n\x13StorageListResponse\x12\x45\n\x08\x65lements\x18\x01 \x03(\x0b\x32).olvid.daemon.datatypes.v1.StorageElementR\x08\x65lements\"%\n\x11StorageGetRequest\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\"*\n\x12StorageGetResponse\x12\x14\n\x05value\x18\x01 \x01(\tR\x05value\";\n\x11StorageSetRequest\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value\";\n\x12StorageSetResponse\x12%\n\x0eprevious_value\x18\x01 \x01(\tR\rpreviousValue\"\'\n\x13StorageUnsetRequest\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\"=\n\x14StorageUnsetResponse\x12%\n\x0eprevious_value\x18\x01 \x01(\tR\rpreviousValue\"\x9c\x01\n\x1c\x44iscussionStorageListRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\x12L\n\x06\x66ilter\x18\x02 \x01(\x0b\x32/.olvid.daemon.datatypes.v1.StorageElementFilterH\x00R\x06\x66ilter\x88\x01\x01\x42\t\n\x07_filter\"f\n\x1d\x44iscussionStorageListResponse\x12\x45\n\x08\x65lements\x18\x01 \x03(\x0b\x32).olvid.daemon.datatypes.v1.StorageElementR\x08\x65lements\"T\n\x1b\x44iscussionStorageGetRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\"4\n\x1c\x44iscussionStorageGetResponse\x12\x14\n\x05value\x18\x01 \x01(\tR\x05value\"j\n\x1b\x44iscussionStorageSetRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\x12\x14\n\x05value\x18\x03 \x01(\tR\x05value\"E\n\x1c\x44iscussionStorageSetResponse\x12%\n\x0eprevious_value\x18\x01 \x01(\tR\rpreviousValue\"V\n\x1d\x44iscussionStorageUnsetRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\"G\n\x1e\x44iscussionStorageUnsetResponse\x12%\n\x0eprevious_value\x18\x01 \x01(\tR\rpreviousValueB\xc3\x01\n\x1b\x63om.olvid.daemon.command.v1B\x14StorageCommandsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODC\xaa\x02\x17Olvid.Daemon.Command.V1\xca\x02\x17Olvid\\Daemon\\Command\\V1\xe2\x02#Olvid\\Daemon\\Command\\V1\\GPBMetadata\xea\x02\x1aOlvid::Daemon::Command::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.command.v1.storage_commands_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.olvid.daemon.command.v1B\024StorageCommandsProtoP\001Z\017olvid.io/daemon\242\002\003ODC\252\002\027Olvid.Daemon.Command.V1\312\002\027Olvid\\Daemon\\Command\\V1\342\002#Olvid\\Daemon\\Command\\V1\\GPBMetadata\352\002\032Olvid::Daemon::Command::V1'
  _globals['_STORAGELISTREQUEST']._serialized_start=116
  _globals['_STORAGELISTREQUEST']._serialized_end=225
  _globals['_STORAGELISTRESPONSE']._serialized_start=227
  _globals['_STORAGELISTRESPONSE']._serialized_end=319
  _globals['_STORAGEGETREQUEST']._serialized_start=321
  _globals['_STORAGEGETREQUEST']._serialized_end=358
  _globals['_STORAGEGETRESPONSE']._serialized_start=360
  _globals['_STORAGEGETRESPONSE']._serialized_end=402
  _globals['_STORAGESETREQUEST']._serialized_start=404
  _globals['_STORAGESETREQUEST']._serialized_end=463
  _globals['_STORAGESETRESPONSE']._serialized_start=465
  _globals['_STORAGESETRESPONSE']._serialized_end=524
  _globals['_STORAGEUNSETREQUEST']._serialized_start=526
  _globals['_STORAGEUNSETREQUEST']._serialized_end=565
  _globals['_STORAGEUNSETRESPONSE']._serialized_start=567
  _globals['_STORAGEUNSETRESPONSE']._serialized_end=628
  _globals['_DISCUSSIONSTORAGELISTREQUEST']._serialized_start=631
  _globals['_DISCUSSIONSTORAGELISTREQUEST']._serialized_end=787
  _globals['_DISCUSSIONSTORAGELISTRESPONSE']._serialized_start=789
  _globals['_DISCUSSIONSTORAGELISTRESPONSE']._serialized_end=891
  _globals['_DISCUSSIONSTORAGEGETREQUEST']._serialized_start=893
  _globals['_DISCUSSIONSTORAGEGETREQUEST']._serialized_end=977
  _globals['_DISCUSSIONSTORAGEGETRESPONSE']._serialized_start=979
  _globals['_DISCUSSIONSTORAGEGETRESPONSE']._serialized_end=1031
  _globals['_DISCUSSIONSTORAGESETREQUEST']._serialized_start=1033
  _globals['_DISCUSSIONSTORAGESETREQUEST']._serialized_end=1139
  _globals['_DISCUSSIONSTORAGESETRESPONSE']._serialized_start=1141
  _globals['_DISCUSSIONSTORAGESETRESPONSE']._serialized_end=1210
  _globals['_DISCUSSIONSTORAGEUNSETREQUEST']._serialized_start=1212
  _globals['_DISCUSSIONSTORAGEUNSETREQUEST']._serialized_end=1298
  _globals['_DISCUSSIONSTORAGEUNSETRESPONSE']._serialized_start=1300
  _globals['_DISCUSSIONSTORAGEUNSETRESPONSE']._serialized_end=1371
# @@protoc_insertion_point(module_scope)



import asyncclick as click

from .tools.click_wrappers import WrapperGroup


@click.group(cls=WrapperGroup)
async def interactive_tree():
	pass


# noinspection PyUnresolvedReferences
from .handler.identity_tree import *
# noinspection PyUnresolvedReferences
from .handler.key_tree import *
# noinspection PyUnresolvedReferences
from .handler.discussion_tree import *
# noinspection PyUnresolvedReferences
from .handler.message_tree import *
# noinspection PyUnresolvedReferences
from .handler.attachment_tree import *
# noinspection PyUnresolvedReferences
from .handler.invitation_tree import *
# noinspection PyUnresolvedReferences
from .handler.contact_tree import *
# noinspection PyUnresolvedReferences
from .handler.group_tree import *
# noinspection PyUnresolvedReferences
from .handler.storage_tree import *
# noinspection PyUnresolvedReferences
from .handler.tools_tree import *



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/notification/v1/group_notifications.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/notification/v1/group_notifications.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import group_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_group__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6olvid/daemon/notification/v1/group_notifications.proto\x12\x1colvid.daemon.notification.v1\x1a%olvid/daemon/datatypes/v1/group.proto\"!\n\x1fSubscribeToGroupNewNotification\"N\n\x14GroupNewNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\"%\n#SubscribeToGroupDeletedNotification\"R\n\x18GroupDeletedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\")\n\'SubscribeToGroupNameUpdatedNotification\"{\n\x1cGroupNameUpdatedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12#\n\rprevious_name\x18\x02 \x01(\tR\x0cpreviousName\"0\n.SubscribeToGroupDescriptionUpdatedNotification\"\x90\x01\n#GroupDescriptionUpdatedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12\x31\n\x14previous_description\x18\x02 \x01(\tR\x13previousDescription\"0\n.SubscribeToGroupPendingMemberAddedNotification\"\xb3\x01\n#GroupPendingMemberAddedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12T\n\x0epending_member\x18\x02 \x01(\x0b\x32-.olvid.daemon.datatypes.v1.PendingGroupMemberR\rpendingMember\"2\n0SubscribeToGroupPendingMemberRemovedNotification\"\xb5\x01\n%GroupPendingMemberRemovedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12T\n\x0epending_member\x18\x02 \x01(\x0b\x32-.olvid.daemon.datatypes.v1.PendingGroupMemberR\rpendingMember\"*\n(SubscribeToGroupMemberJoinedNotification\"\x97\x01\n\x1dGroupMemberJoinedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12>\n\x06member\x18\x02 \x01(\x0b\x32&.olvid.daemon.datatypes.v1.GroupMemberR\x06member\"(\n&SubscribeToGroupMemberLeftNotification\"\x95\x01\n\x1bGroupMemberLeftNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12>\n\x06member\x18\x02 \x01(\x0b\x32&.olvid.daemon.datatypes.v1.GroupMemberR\x06member\"3\n1SubscribeToGroupOwnPermissionsUpdatedNotification\"\x9b\x02\n&GroupOwnPermissionsUpdatedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12S\n\x0bpermissions\x18\x02 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x0bpermissions\x12\x64\n\x14previous_permissions\x18\x03 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x13previousPermissions\"6\n4SubscribeToGroupMemberPermissionsUpdatedNotification\"\x89\x02\n)GroupMemberPermissionsUpdatedNotification\x12\x36\n\x05group\x18\x01 \x01(\x0b\x32 .olvid.daemon.datatypes.v1.GroupR\x05group\x12>\n\x06member\x18\x02 \x01(\x0b\x32&.olvid.daemon.datatypes.v1.GroupMemberR\x06member\x12\x64\n\x14previous_permissions\x18\x03 \x01(\x0b\x32\x31.olvid.daemon.datatypes.v1.GroupMemberPermissionsR\x13previousPermissions\".\n,SubscribeToGroupUpdateInProgressNotification\">\n!GroupUpdateInProgressNotification\x12\x19\n\x08group_id\x18\x01 \x01(\x04R\x07groupId\",\n*SubscribeToGroupUpdateFinishedNotification\"<\n\x1fGroupUpdateFinishedNotification\x12\x19\n\x08group_id\x18\x01 \x01(\x04R\x07groupIdB\xdf\x01\n com.olvid.daemon.notification.v1B\x17GroupNotificationsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODN\xaa\x02\x1cOlvid.Daemon.Notification.V1\xca\x02\x1cOlvid\\Daemon\\Notification\\V1\xe2\x02(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\xea\x02\x1fOlvid::Daemon::Notification::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.notification.v1.group_notifications_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n com.olvid.daemon.notification.v1B\027GroupNotificationsProtoP\001Z\017olvid.io/daemon\242\002\003ODN\252\002\034Olvid.Daemon.Notification.V1\312\002\034Olvid\\Daemon\\Notification\\V1\342\002(Olvid\\Daemon\\Notification\\V1\\GPBMetadata\352\002\037Olvid::Daemon::Notification::V1'
  _globals['_SUBSCRIBETOGROUPNEWNOTIFICATION']._serialized_start=127
  _globals['_SUBSCRIBETOGROUPNEWNOTIFICATION']._serialized_end=160
  _globals['_GROUPNEWNOTIFICATION']._serialized_start=162
  _globals['_GROUPNEWNOTIFICATION']._serialized_end=240
  _globals['_SUBSCRIBETOGROUPDELETEDNOTIFICATION']._serialized_start=242
  _globals['_SUBSCRIBETOGROUPDELETEDNOTIFICATION']._serialized_end=279
  _globals['_GROUPDELETEDNOTIFICATION']._serialized_start=281
  _globals['_GROUPDELETEDNOTIFICATION']._serialized_end=363
  _globals['_SUBSCRIBETOGROUPNAMEUPDATEDNOTIFICATION']._serialized_start=365
  _globals['_SUBSCRIBETOGROUPNAMEUPDATEDNOTIFICATION']._serialized_end=406
  _globals['_GROUPNAMEUPDATEDNOTIFICATION']._serialized_start=408
  _globals['_GROUPNAMEUPDATEDNOTIFICATION']._serialized_end=531
  _globals['_SUBSCRIBETOGROUPDESCRIPTIONUPDATEDNOTIFICATION']._serialized_start=533
  _globals['_SUBSCRIBETOGROUPDESCRIPTIONUPDATEDNOTIFICATION']._serialized_end=581
  _globals['_GROUPDESCRIPTIONUPDATEDNOTIFICATION']._serialized_start=584
  _globals['_GROUPDESCRIPTIONUPDATEDNOTIFICATION']._serialized_end=728
  _globals['_SUBSCRIBETOGROUPPENDINGMEMBERADDEDNOTIFICATION']._serialized_start=730
  _globals['_SUBSCRIBETOGROUPPENDINGMEMBERADDEDNOTIFICATION']._serialized_end=778
  _globals['_GROUPPENDINGMEMBERADDEDNOTIFICATION']._serialized_start=781
  _globals['_GROUPPENDINGMEMBERADDEDNOTIFICATION']._serialized_end=960
  _globals['_SUBSCRIBETOGROUPPENDINGMEMBERREMOVEDNOTIFICATION']._serialized_start=962
  _globals['_SUBSCRIBETOGROUPPENDINGMEMBERREMOVEDNOTIFICATION']._serialized_end=1012
  _globals['_GROUPPENDINGMEMBERREMOVEDNOTIFICATION']._serialized_start=1015
  _globals['_GROUPPENDINGMEMBERREMOVEDNOTIFICATION']._serialized_end=1196
  _globals['_SUBSCRIBETOGROUPMEMBERJOINEDNOTIFICATION']._serialized_start=1198
  _globals['_SUBSCRIBETOGROUPMEMBERJOINEDNOTIFICATION']._serialized_end=1240
  _globals['_GROUPMEMBERJOINEDNOTIFICATION']._serialized_start=1243
  _globals['_GROUPMEMBERJOINEDNOTIFICATION']._serialized_end=1394
  _globals['_SUBSCRIBETOGROUPMEMBERLEFTNOTIFICATION']._serialized_start=1396
  _globals['_SUBSCRIBETOGROUPMEMBERLEFTNOTIFICATION']._serialized_end=1436
  _globals['_GROUPMEMBERLEFTNOTIFICATION']._serialized_start=1439
  _globals['_GROUPMEMBERLEFTNOTIFICATION']._serialized_end=1588
  _globals['_SUBSCRIBETOGROUPOWNPERMISSIONSUPDATEDNOTIFICATION']._serialized_start=1590
  _globals['_SUBSCRIBETOGROUPOWNPERMISSIONSUPDATEDNOTIFICATION']._serialized_end=1641
  _globals['_GROUPOWNPERMISSIONSUPDATEDNOTIFICATION']._serialized_start=1644
  _globals['_GROUPOWNPERMISSIONSUPDATEDNOTIFICATION']._serialized_end=1927
  _globals['_SUBSCRIBETOGROUPMEMBERPERMISSIONSUPDATEDNOTIFICATION']._serialized_start=1929
  _globals['_SUBSCRIBETOGROUPMEMBERPERMISSIONSUPDATEDNOTIFICATION']._serialized_end=1983
  _globals['_GROUPMEMBERPERMISSIONSUPDATEDNOTIFICATION']._serialized_start=1986
  _globals['_GROUPMEMBERPERMISSIONSUPDATEDNOTIFICATION']._serialized_end=2251
  _globals['_SUBSCRIBETOGROUPUPDATEINPROGRESSNOTIFICATION']._serialized_start=2253
  _globals['_SUBSCRIBETOGROUPUPDATEINPROGRESSNOTIFICATION']._serialized_end=2299
  _globals['_GROUPUPDATEINPROGRESSNOTIFICATION']._serialized_start=2301
  _globals['_GROUPUPDATEINPROGRESSNOTIFICATION']._serialized_end=2363
  _globals['_SUBSCRIBETOGROUPUPDATEFINISHEDNOTIFICATION']._serialized_start=2365
  _globals['_SUBSCRIBETOGROUPUPDATEFINISHEDNOTIFICATION']._serialized_end=2409
  _globals['_GROUPUPDATEFINISHEDNOTIFICATION']._serialized_start=2411
  _globals['_GROUPUPDATEFINISHEDNOTIFICATION']._serialized_end=2471
# @@protoc_insertion_point(module_scope)



from google.protobuf.json_format import Parse, ParseError

from ..interactive_tree import interactive_tree
from ..tools.cli_tools import *
from ..tools.click_wrappers import WrapperGroup


#####
# discussion
#####
@interactive_tree.group("discussion", help="manage your discussions", cls=WrapperGroup)
def discussion_tree():
	pass


#####
# discussion get
#####
# noinspection PyProtectedMember
@discussion_tree.command("get", help="list current identity discussions")
@click.option("-a", "--all", "get_all", is_flag=True)
@click.option("-c", "--contact", "by_contact", is_flag=True)
@click.option("-g", "--group", "by_group", is_flag=True)
@click.argument("discussion_ids", nargs=-1, type=click.INT)
@click.option("-f", "--fields", "fields", type=str)
@click.option("--filter", "filter_", type=str)
async def discussion_get(get_all, by_contact: bool, by_group: bool, discussion_ids, fields: str, filter_: str = ""):
	# build filter
	discussion_filter: datatypes.DiscussionFilter = datatypes.DiscussionFilter()
	if filter_:
		try:
			parsed_message = Parse(filter_, datatypes.DiscussionFilter()._to_native(discussion_filter))
			discussion_filter = datatypes.DiscussionFilter._from_native(parsed_message)
		except ParseError as e:
			print_error_message(f"Cannot parse filter: {e}")
			return

	discussions: list[datatypes.Discussion]
	if get_all or not discussion_ids:
		discussions = [d async for d in ClientSingleton.get_client().discussion_list(filter=discussion_filter)]
	elif by_contact:
		discussions = [await ClientSingleton.get_client().discussion_get_by_contact(did) for did in discussion_ids]
	elif by_group:
		discussions = [await ClientSingleton.get_client().discussion_get_by_group(did) for did in discussion_ids]
	else:
		discussions = [await ClientSingleton.get_client().discussion_get(did) for did in discussion_ids]
	for discussion in discussions:
		filter_fields_and_print_normal_message(discussion, fields)


#####
# discussion rm
#####
@discussion_tree.command("empty", help="delete all messages in a discussion")
@click.argument("discussion_ids", nargs=-1, type=click.INT, required=True)
@click.option("-e", "--everywhere", "delete_everywhere", is_flag=True, default=False)
async def discussion_rm(discussion_ids: tuple[int], delete_everywhere: bool = False):
	for discussion_id in discussion_ids:
		await ClientSingleton.get_client().discussion_empty(discussion_id=discussion_id,
															delete_everywhere=delete_everywhere)
		print_command_result(f"Discussion emptied: {discussion_id}")


#####
# discussion settings
#####
@discussion_tree.group("settings", help="manage discussion settings", cls=WrapperGroup)
def settings_tree():
	pass


#####
# discussion settings get
#####
@settings_tree.command("get", help="get a ")
@click.argument("discussion_id", nargs=1, type=click.INT, required=True)
@click.option("-f", "--fields", "fields", type=str)
async def discussion_settings_get(discussion_id: int, fields: str):
	settings = await ClientSingleton.get_client().discussion_settings_get(discussion_id=discussion_id)
	filter_fields_and_print_normal_message(settings, fields)


#####
# discussion settings set
#####
@settings_tree.command("set")
@click.option("-o", "--once", "read_once", is_flag=True)
@click.option("-e", "--existence", "existence_duration", type=click.INT, default=0)
@click.option("-v", "--visibility", "visibility_duration", type=click.INT, default=0)
@click.argument("discussion_id", nargs=1, type=click.INT)
async def discussion_settings_set(discussion_id: int, read_once: bool, existence_duration: int,
									visibility_duration: int):
	settings = datatypes.DiscussionSettings(discussion_id=discussion_id,
											read_once=read_once,
											existence_duration=existence_duration,
											visibility_duration=visibility_duration)
	new_settings = await ClientSingleton.get_client().discussion_settings_set(settings=settings)
	print_normal_message(new_settings, new_settings)


#####
# discussion locked
#####
@discussion_tree.group("locked", help="manage locked discussion", cls=WrapperGroup)
def locked_tree():
	pass


#####
# discussion locked get
#####
@locked_tree.command("get", help="list locked discussions")
@click.option("-f", "--fields", "fields", type=str)
async def discussion_locked_get(fields: str):
	async for discussion in ClientSingleton.get_client().discussion_locked_list():
		filter_fields_and_print_normal_message(discussion, fields)


#####
# discussion locked rm
#####
@locked_tree.command("rm", help="delete locked discussion")
@click.argument("discussion_ids", nargs=-1, type=click.INT, required=True)
async def discussion_rm(discussion_ids: tuple[int]):
	for discussion_id in discussion_ids:
		await ClientSingleton.get_client().discussion_locked_delete(discussion_id=discussion_id)
		print_command_result(f"Locked discussion deleted: {discussion_id}")



from setuptools import setup, find_packages
from setuptools import Command


# read version from version.py file in olvid module
version = {}
with open("olvid/version.py") as fd:
    exec(fd.read().strip(), version)


class VersionCommand(Command):
    description = "Print olvid module version"
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version['__version__'])


setup(
    name='olvid-bot',
    version=version["__version__"],
    description="Module to communicate with Olvid daemon project",
    author="olvid.io",
    author_email="bot@olvid.io",
    packages=find_packages(
        include=[
            "olvid",
            "olvid.cli", "olvid.cli.handler", "olvid.cli.tools",
            "olvid.core",
            "olvid.listeners",
            "olvid.datatypes",
            "olvid.internal", "olvid.internal.admin", "olvid.internal.commands", "olvid.internal.notifications", "olvid.internal.types",
            "olvid.tools"
        ]
    ) + [
        "olvid.protobuf", "olvid.protobuf.olvid", "olvid.protobuf.olvid.daemon",
        "olvid.protobuf.olvid.daemon.services", "olvid.protobuf.olvid.daemon.services.v1",
        "olvid.protobuf.olvid.daemon.admin", "olvid.protobuf.olvid.daemon.admin.v1",
        "olvid.protobuf.olvid.daemon.command", "olvid.protobuf.olvid.daemon.command.v1",
        "olvid.protobuf.olvid.daemon.datatypes", "olvid.protobuf.olvid.daemon.datatypes.v1",
        "olvid.protobuf.olvid.daemon.notification", "olvid.protobuf.olvid.daemon.notification.v1"
    ],
    package_data={'olvid.core.calls.aiortc.codecs': ['*.pyi', "*.so"]},
    include_package_data=True,
    cmdclass={
        "version": VersionCommand,
    },
)



import asyncio
import inspect
import json
from abc import ABC, abstractmethod
from asyncio import Task
from datetime import datetime
from typing import Optional, Coroutine

from olvid import OlvidClient

from .logger import tools_logger


# the class that will manage tasks have to extend this class.
# - storage_prefix: must be unique for every task holder.
#   this prefix allows holder to recognize its fields in the database.
# Initialization: actually self.load method is called asynchronously, in the background.
# so this class might not be initialized on the instance creation.
# if you need to know when it's initialized, you can check self._tasks_map is not None.
# if you need to force initialization, you can call self.load by yourself
class DelayedTasksHolder:
	# override this abstract class to use the DelayedTaskHolder
	# Instances will be json serialized, so you must store necessary data in the dict returned by serialize method.
	# - discussion_id is a very important attribute if you are creating tasks associated with a discussion,
	# 	and you want these tasks to be still associated with this discussion after a backup restoration.
	#   if you don't use discussion, you can set discussion_id to 0 all the time
	# - task_id must be unique.
	# if you need to update / delete your tasks, you will need to guess this id.
	# in that case, most of the time,
	#   use the message_id that created this task.
	#   if you don't need to retrieve your task later this id can be anything random.
	class DelayedTask(ABC):
		@abstractmethod
		def get_execution_timestamp(self) -> int:
			pass

		@abstractmethod
		def get_task_id(self) -> str:
			pass

		@abstractmethod
		def get_discussion_id(self) -> Optional[int]:
			pass

		@abstractmethod
		def is_repeated(self) -> bool:
			pass

		@abstractmethod
		def serialize(self) -> dict:
			pass

		@staticmethod
		@abstractmethod
		def deserialize(serialized_dict: dict) -> "DelayedTasksHolder.DelayedTask":
			pass

		@abstractmethod
		async def run(self, client: OlvidClient):
			pass

	# Storage_prefix must be a unique and constant string.
	# It's used to determine if an element in storage is associated with a specific task holder.
	def __init__(self, client: OlvidClient, task_class: type[DelayedTask], storage_prefix: str):
		# we use olvid client to access storage, and we pass it to tasks on execution.
		self._client: OlvidClient = client
		# we use a unique storage prefix to easily determine tasks associated with a task holder in daemon storage.
		# this prefix must be uniquely associated with a task holder class.
		self._storage_prefix: str = storage_prefix
		# We need task class to deserialize tasks
		self._task_class: type[DelayedTasksHolder.DelayedTask] = task_class

		# we keep in memory a copy of storage state to make next timer computation easier.
		# this map is set to None until load method is called and initialize it
		self._tasks_map: Optional[dict[str, DelayedTasksHolder.DelayedTask]] = None

		# store current timer task (task waiting for the next execution and that will execute passed events handlers)
		self._timer_task: Optional[asyncio.Task] = None
		# store a datetime object holding the next timer_task expiration
		self._current_timer_execution_timestamp: Optional[int] = None
		# A set of all background tasks.
		# We use it to execute tasks in the background.
		# Tasks are automatically removed when executed.
		self.__task_set = set()

		# load registered tasks lazily because we must do asyncio calls (the task map will be None until the loading end)
		self._add_background_task(self.load(), f"{self.__class__.__name__}-load-on-start")

	# load previous tasks
	async def load(self):
		if self._tasks_map is not None:
			return
		# this instance has already been initialized
		self._tasks_map = {}
		# load global tasks
		async for element in self._client.storage_list():
			if self._is_valid_task_storage_key(key=element.key):
				try:
					task = self._task_class.deserialize(json.loads(element.value))
					self._tasks_map[task.get_task_id()] = task
					tools_logger.info(f"Loaded task: {task.get_task_id()}: {task.get_execution_timestamp()}")
				except json.JSONDecodeError:
					tools_logger.exception(f"{self.__class__.__name__}: list_tasks: discussion ({0}): cannot decode element in storage with key: {element.key}")
					continue

		# load discussion tasks
		async for discussion in self._client.discussion_list():
			async for element in self._client.discussion_storage_list(discussion_id=discussion.id):
				if element.key.startswith(self._storage_prefix):
					try:
						task = self._task_class.deserialize(json.loads(element.value))
						self._tasks_map[task.get_task_id()] = task
						tools_logger.info(f"Loaded task: {task.get_task_id()}: {task.get_execution_timestamp()}")
					except Exception:
						tools_logger.exception(f"{self.__class__.__name__}: list_tasks: discussion ({discussion.id}): cannot decode element in storage with key: {element.key}")

		self._refresh_timer()

	####
	# Public Api
	####
	# set discussion_id to 0 if you want to list tasks not associated to a discussion
	async def list_tasks(self, discussion_id: int) -> list["DelayedTasksHolder.DelayedTask"]:
		return [task for task in self._tasks_map.values() if task.get_discussion_id() == discussion_id]

	# set discussion_id to 0 if you want to get a task not associated to a discussion
	async def get_task(self, task_id: str, discussion_id: int) -> Optional["DelayedTasksHolder.DelayedTask"]:
		task = self._tasks_map.get(task_id)
		return None if not task else task if task.get_discussion_id() == discussion_id else None

	async def add_task(self, task: "DelayedTasksHolder.DelayedTask"):
		# update storage
		task_storage_key = self._get_task_storage_key(task.get_task_id())
		if not task.get_discussion_id():
			await self._client.storage_set(key=task_storage_key, value=json.dumps(task.serialize()))
		else:
			await self._client.discussion_storage_set(discussion_id=task.get_discussion_id(), key=task_storage_key, value=json.dumps(task.serialize()))

		# update task map
		self._tasks_map[task.get_task_id()] = task

		# refresh timer
		self._refresh_timer()

	async def remove_task(self, task: "DelayedTasksHolder.DelayedTask"):
		# update storage
		if task.get_discussion_id() == 0:
			await self._client.storage_unset(key=self._get_task_storage_key(task.get_task_id()))
		else:
			await self._client.discussion_storage_unset(discussion_id=task.get_discussion_id(), key=self._get_task_storage_key(task.get_task_id()))

		# update task map
		self._tasks_map.pop(task.get_task_id())

		# refresh timer
		self._refresh_timer()

	async def update_task(self, updated_task: "DelayedTasksHolder.DelayedTask"):
		# update storage
		task_storage_key = self._get_task_storage_key(updated_task.get_task_id())
		if updated_task.get_discussion_id() == 0:
			await self._client.storage_set(key=task_storage_key, value=json.dumps(updated_task.serialize()))
		else:
			await self._client.discussion_storage_set(discussion_id=updated_task.get_discussion_id(), key=task_storage_key, value=json.dumps(updated_task.serialize()))

		# update task map
		self._tasks_map[updated_task.get_task_id()] = updated_task

		# refresh timer
		self._refresh_timer()

	####
	# Timer api
	####
	def _refresh_timer(self):
		try:
			# no more tasks to handle
			if len(self._tasks_map) == 0:
				if self._timer_task:
					self._timer_task.cancel()
					self._timer_task = None
				return

			# determine next task to execute
			next_task: Optional[DelayedTasksHolder.DelayedTask] = None
			next_execution: int = self._current_timer_execution_timestamp
			for task in self._tasks_map.values():
				if next_execution is None:
					next_execution = task.get_execution_timestamp()
					next_task = task
				elif task.get_execution_timestamp() < next_execution:
					next_execution = task.get_execution_timestamp()
					next_task = task

			# if next handler is set this means that timer delay changed and it must be updated
			if next_task is not None:
				# stop previous background timer task
				if self._timer_task:
					self._timer_task.cancel()

				# setup timer
				delay_s = next_execution - datetime.now().timestamp()
				# if delay is negative, some tasks require to be executed now (timer_task will re-sync timer)
				if delay_s < 0:
					self._timer_task = self._add_background_task(self._timer_task_method(0), name=f"{self.__class__.__name__}-timer-task")
					return
				# else setup a timer
				self._timer_task = self._add_background_task(self._timer_task_method(delay_s), name=f"{self.__class__.__name__}-timer-task")
				self._current_timer_execution_timestamp = next_execution
		except Exception:
			tools_logger.exception("Unexpected exception in _refresh_timer")

	async def _timer_task_method(self, delay_s: float):		# await for delay
		try:
			if delay_s > 0:
				await asyncio.sleep(delay_s)

			# executed expired tasks
			tasks_to_remove: list[DelayedTasksHolder.DelayedTask] = []
			now = datetime.now()

			for task_id, task in self._tasks_map.items():
				try:
					if task.get_execution_timestamp() <= now.timestamp():
						ret = task.run(self._client)
						if inspect.iscoroutine(ret):
							self._add_background_task(ret, name=f"{self.__class__.__name__}-{task_id}")
						tasks_to_remove.append(task)
				except Exception:
					tools_logger.exception(f"{self.__class__.__name__}: _timer_task_method: an error occurred when running a task: {task_id}")

			# if ran tasks are not repeated, we can delete them
			for task in tasks_to_remove:
				if task.is_repeated():
					# update task in memory to update next execution timestamp
					await self.update_task(task)
				else:
					await self.remove_task(task)

			# reset timer delay
			self._current_timer_execution_timestamp = None

			# relaunch timer
			self._refresh_timer()
		except asyncio.CancelledError:
			raise
		except Exception:
			tools_logger.exception("Unexpected exception in _time_task_method")

	####
	# Internal
	####
	def _get_task_storage_key(self, task_id: str) -> str:
		return f"{self._storage_prefix}{task_id}"

	def _is_valid_task_storage_key(self, key: str) -> bool:
		return key.startswith(self._storage_prefix) and len(key) > len(self._storage_prefix)

	def _add_background_task(self, coroutine: Coroutine, name: str = "") -> Task:
		task = asyncio.get_event_loop().create_task(coroutine, name=name if name else None)
		self.__task_set.add(task)

		def end_callback(t):
			self.__task_set.remove(t)
		task.add_done_callback(end_callback)
		return task



import copy
import logging
import traceback
from typing import Callable

import grpc.aio

from olvid import OlvidBot, datatypes, listeners

from ClientHolder import ClientHolder, ClientWrapper


# noinspection DuplicatedCode
async def generic_read_once_routine(client_1: ClientWrapper, client_2: ClientWrapper, test_id: str, ephemerality: datatypes.MessageEphemerality):
	discussion_1, discussion_2 = await client_1.get_discussion_associated_to_another_client(client_2), await client_2.get_discussion_associated_to_another_client(client_1)
	contact_2 = await client_2.get_contact_associated_to_another_client(client_1)

	####
	# read once messages
	####
	bot_1_ideal_message = datatypes.Message(id=datatypes.MessageId(type=datatypes.MessageId.Type.TYPE_OUTBOUND), discussion_id=discussion_1.id, body=test_id)
	bot_2_ideal_message = datatypes.Message(id=datatypes.MessageId(type=datatypes.MessageId.Type.TYPE_INBOUND), discussion_id=discussion_2.id, body=test_id, sender_id=contact_2.id)

	bots: list[OlvidBot] = []
	optional_bots: list[OlvidBot] = []
	# check notification flow on client 1
	bots.append(client_1.create_notification_bot(listeners.MessageSentListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_SENT), count=1)))
	bots.append(client_1.create_notification_bot(listeners.MessageDeletedListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELETED), count=1)))

	optional_bots.append(client_1.create_notification_bot(listeners.MessageUploadedListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_UPLOADED), count=1)))
	optional_bots.append(client_1.create_notification_bot(listeners.MessageDeliveredListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELIVERED), count=1)))
	optional_bots.append(client_1.create_notification_bot(listeners.MessageReadListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_READ), count=1)))

	# check notification flow on client 2
	bots.append(client_2.create_notification_bot(listeners.MessageReceivedListener(handler=client_2.get_check_content_handler(bot_2_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_RECEIVED), count=1)))
	bots.append(client_2.create_notification_bot(listeners.MessageDeletedListener(handler=client_2.get_check_content_handler(bot_2_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELETED), count=1)))

	# check message deletion on both clients
	bots.append(client_1.create_notification_bot(listeners.MessageDeletedListener(handler=assert_message_deletion_handler_creator(client_1), count=1)))
	bots.append(client_2.create_notification_bot(listeners.MessageDeletedListener(handler=assert_message_deletion_handler_creator(client_2), count=1)))

	# send message
	message = await client_1.message_send(discussion_id=discussion_1.id, body=test_id, ephemerality=ephemerality)
	logging.debug(f"Sent {test_id} message: {message_id_to_str(message.id)}")

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()
	for bot in optional_bots:
		await bot.stop()


# noinspection DuplicatedCode
async def generic_self_expiring_test(client_1: ClientWrapper, client_2: ClientWrapper, test_id: str, ephemerality: datatypes.MessageEphemerality):
	discussion_1, discussion_2 = await client_1.get_discussion_associated_to_another_client(client_2), await client_2.get_discussion_associated_to_another_client(client_1)
	contact_2 = await client_2.get_contact_associated_to_another_client(client_1)

	####
	# existence duration messages
	####
	bot_1_ideal_message = datatypes.Message(id=datatypes.MessageId(type=datatypes.MessageId.Type.TYPE_OUTBOUND), discussion_id=discussion_1.id, body=test_id)
	bot_2_ideal_message = datatypes.Message(id=datatypes.MessageId(type=datatypes.MessageId.Type.TYPE_INBOUND), discussion_id=discussion_2.id, body=test_id, sender_id=contact_2.id)

	bots: list[OlvidBot] = [
		# check notification flow on client 1
		client_1.create_notification_bot(listeners.MessageSentListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_SENT), count=1)),
		client_1.create_notification_bot(listeners.MessageUploadedListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_UPLOADED), count=1)),
		client_1.create_notification_bot(listeners.MessageDeliveredListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELIVERED), count=1)),
		client_1.create_notification_bot(listeners.MessageReadListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_READ), count=1)),
		client_1.create_notification_bot(listeners.MessageDeletedListener(handler=client_1.get_check_content_handler(bot_1_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELETED), count=1)),

		# check notification flow on client 2
		client_2.create_notification_bot(listeners.MessageReceivedListener(handler=client_2.get_check_content_handler(bot_2_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_RECEIVED), count=1)),
		client_2.create_notification_bot(listeners.MessageDeletedListener(handler=client_2.get_check_content_handler(bot_2_ideal_message, notification_type=listeners.NOTIFICATIONS.MESSAGE_DELETED), count=1)),

		# check message is available on both clients, on every notification (before deletion)
		client_1.create_notification_bot(listeners.MessageSentListener(handler=assert_message_existence_handler_creator(client_1, handler_id=f"{test_id}-4.1"), count=1)),
		client_1.create_notification_bot(listeners.MessageUploadedListener(handler=assert_message_existence_handler_creator(client_1, handler_id=f"{test_id}-4.2"), count=1)),
		client_1.create_notification_bot(listeners.MessageDeliveredListener(handler=assert_message_existence_handler_creator(client_1, handler_id=f"{test_id}-4.3"), count=1)),
		client_1.create_notification_bot(listeners.MessageReadListener(handler=assert_message_existence_handler_creator(client_1, handler_id=f"{test_id}-4.3"), count=1)),
		client_2.create_notification_bot(listeners.MessageReceivedListener(handler=assert_message_existence_handler_creator(client_2, handler_id=f"{test_id}-4.4"), count=1)),

		# check message had been deleted on both client when delete notifications arrive
		client_1.create_notification_bot(listeners.MessageDeletedListener(handler=assert_message_deletion_handler_creator(client_1, handler_id=f"{test_id}-3.1"), count=1)),
		client_2.create_notification_bot(listeners.MessageDeletedListener(handler=assert_message_deletion_handler_creator(client_2, handler_id=f"{test_id}-3.2"), count=1)),

		# check message deletion after the delay
		# client_1.create_notification_bot(notifications_dictionary={Notifications.MESSAGE_DELIVERED.name: 1}, handler=assert_message_deletion_handler_creator(client_1, delay=delay_before_deletion_check + 1)),
		# client_2.create_notification_bot(notifications_dictionary={Notifications.MESSAGE_RECEIVED.name: 1}, handler=assert_message_deletion_handler_creator(client_2, delay=delay_before_deletion_check + 1)),

	]

	# send message
	message = await client_1.message_send(discussion_id=discussion_1.id, body=test_id, ephemerality=ephemerality)
	logging.debug(f"Sent {test_id} message: {message_id_to_str(message.id)}")

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()


# execute generic ephemeral tests, but it set discussion shared ephemeral settings before test (and restore it after)
# noinspection DuplicatedCode
async def generic_discussion_customization_test(client_1: ClientWrapper, client_2: ClientWrapper, test_id: str, new_settings: datatypes.DiscussionSettings, generic_routine: Callable):
	logging.debug(f"Starting test: {test_id}")

	discussion_1, discussion_2 = await client_1.get_discussion_associated_to_another_client(client_2), await client_2.get_discussion_associated_to_another_client(client_1)

	# get current settings
	prev_settings_1 = (await client_1.discussion_settings_get(discussion_1.id))
	prev_settings_2 = (await client_2.discussion_settings_get(discussion_2.id))

	# prepare new settings
	new_settings_1 = copy.deepcopy(new_settings)
	new_settings_1.discussion_id = prev_settings_1.discussion_id
	new_settings_2 = copy.deepcopy(new_settings)
	new_settings_2.discussion_id = prev_settings_2.discussion_id

	# prepare notification bots
	bot_1 = client_1.create_notification_bot(listeners.DiscussionSettingsUpdatedListener(handler=client_1.get_check_content_handler(new_settings_1, prev_settings_1, notification_type=listeners.NOTIFICATIONS.DISCUSSION_SETTINGS_UPDATED), count=1))
	bot_2 = client_2.create_notification_bot(listeners.DiscussionSettingsUpdatedListener(handler=client_2.get_check_content_handler(new_settings_2, prev_settings_2, notification_type=listeners.NOTIFICATIONS.DISCUSSION_SETTINGS_UPDATED), count=1))

	# set ephemeral settings
	await client_1.discussion_settings_set(settings=new_settings_1)

	# wait for notifications to arrive
	await bot_1.wait_for_listeners_end()
	await bot_2.wait_for_listeners_end()

	# test settings are properly used
	await generic_routine(client_1, client_2, test_id, datatypes.MessageEphemerality())

	# restore previous settings
	# prepare notification bots
	bot_1 = client_1.create_notification_bot(listeners.DiscussionSettingsUpdatedListener(handler=client_1.get_check_content_handler(prev_settings_1, new_settings_1, notification_type=listeners.NOTIFICATIONS.DISCUSSION_SETTINGS_UPDATED), count=1))
	bot_2 = client_2.create_notification_bot(listeners.DiscussionSettingsUpdatedListener(handler=client_2.get_check_content_handler(prev_settings_2, new_settings_2, notification_type=listeners.NOTIFICATIONS.DISCUSSION_SETTINGS_UPDATED), count=1))

	# set ephemeral settings
	await client_2.discussion_settings_set(settings=prev_settings_2)

	# wait for notifications to arrive
	await bot_1.wait_for_listeners_end()
	await bot_2.wait_for_listeners_end()


def assert_message_deletion_handler_creator(client: ClientWrapper, handler_id: str = ""):
	async def assert_message_deletion_handler(message: datatypes.Message):
		try:
			message_id = message.id
			try:
				await client.message_get(message_id=message_id)
				raise Exception("Message was not deleted on server")
			except grpc.aio.AioRpcError:
				pass

			async for message in client.message_list():
				assert message.id != message_id
			logging.debug(f"Message was deleted on server: {message_id_to_str(message_id)}")
		except Exception as e:
			logging.error(f"{handler_id}: exception")
			traceback.print_exc()
			raise e
	return assert_message_deletion_handler


def assert_message_existence_handler_creator(client: ClientWrapper, handler_id: str = ""):
	async def assert_message_existence_handler(message: datatypes.Message):
		try:
			message_id = message.id
			await client.message_get(message_id=message_id)

			async for message in client.message_list():
				if message.id == message_id:
					logging.debug(f"Message was found on server: {message_id_to_str(message_id)}")
					return
			raise Exception("Message not found using message_list")
		except Exception as e:
			logging.error(f"{handler_id}: exception")
			traceback.print_exc()
			raise e
	return assert_message_existence_handler


def message_id_to_str(message_id: datatypes.MessageId) -> str:
	return f"{message_id.type.name[0]}-{message_id.id}"


async def test_ephemeral_message(client_holder: ClientHolder, fast_mode: bool = False):
	for c1, c2 in client_holder.get_all_client_pairs():
		logging.info("-" * 10 + " Read Once " + "-" * 10)
		await generic_read_once_routine(c1, c2, "read-once", datatypes.MessageEphemerality(read_once=True))

		logging.info("-" * 10 + " Visibility Duration " + "-" * 10)
		await generic_self_expiring_test(c1, c2, "visibility", datatypes.MessageEphemerality(visibility_duration=3))

		logging.info("-" * 10 + " Existence Duration " + "-" * 10)
		await generic_self_expiring_test(c1, c2, "existence", datatypes.MessageEphemerality(existence_duration=4))

		logging.info("-" * 10 + " Visibility-Existence Duration " + "-" * 10)
		await generic_self_expiring_test(c1, c2, "vi-ex", datatypes.MessageEphemerality(existence_duration=20, visibility_duration=3))

		logging.info("-" * 10 + " Existence-Visibility Duration " + "-" * 10)
		await generic_self_expiring_test(c1, c2, "ex-vi", datatypes.MessageEphemerality(existence_duration=4, visibility_duration=20))

		logging.info("-" * 10 + " ReadOnce-Existence-Visibility Duration " + "-" * 10)
		await generic_read_once_routine(c1, c2, "ro-ex-vi", datatypes.MessageEphemerality(read_once=True, existence_duration=3, visibility_duration=2))

		logging.info("-" * 10 + " DiscussionCustomization-1 " + "-" * 10)
		await generic_discussion_customization_test(c1, c2, "dc-1", datatypes.DiscussionSettings(read_once=True), generic_read_once_routine)

		if not fast_mode:
			logging.info("-" * 10 + " DiscussionCustomization-2 " + "-" * 10)
			await generic_discussion_customization_test(c1, c2, "dc-2", datatypes.DiscussionSettings(visibility_duration=3), generic_self_expiring_test)

			logging.info("-" * 10 + " DiscussionCustomization-3 " + "-" * 10)
			await generic_discussion_customization_test(c1, c2, "dc-3", datatypes.DiscussionSettings(existence_duration=4, visibility_duration=20), generic_self_expiring_test)

		if fast_mode:
			break



import asyncio

import grpc.aio
from google.protobuf.json_format import Parse, ParseError

from ..interactive_tree import interactive_tree
from ..tools.cli_tools import *
from olvid import datatypes
from ..tools.click_wrappers import WrapperGroup
from ..tools.interactive_actions import ask_question_with_context, contact_new, print_with_context


#####
# identity
#####
@interactive_tree.group("identity", help="select current identity, (admin) manage olvid identities", cls=WrapperGroup)
def identity_tree():
	pass


#####
# identity current
#####
@identity_tree.command("current", help="get or update identity currently used for your commands")
@click.argument("identity_id", nargs=1, type=click.INT, required=False)
async def identity_current(identity_id: int):
	# If no arguments show current identity
	if (not identity_id):
		if not ClientSingleton.get_current_identity_id():
			print_command_result("No current identity")
		else:
			identity = await ClientSingleton.get_client().identity_get()
			print_command_result("Identity currently used: {}".format(identity.details))
		return

	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot change current identity when impersonating a non admin key")
		return

	# unset current identity
	if identity_id == 0:
		ClientSingleton.set_current_identity_id(0)
		print_command_result("Unselected current identity")
		return

	# change current identity index (save previous version)
	previous_identity_id = ClientSingleton.get_current_identity_id()
	ClientSingleton.set_current_identity_id(identity_id)

	try:
		# check if identity exists
		identity = await ClientSingleton.get_client().identity_get()
		# hide invitation url and show result
		identity.invitation_url = ""
		print_normal_message(identity, identity.id)
	except grpc.aio.AioRpcError as e:
		# identity does not exist, restore previous id
		if e.code() == grpc.StatusCode.NOT_FOUND:
			ClientSingleton.set_current_identity_id(previous_identity_id)
		raise e


#####
# identity new
#####
@identity_tree.command("new", help="create a new identity")
@click.option("-f", "--first", "first_opt", default="", help="first name")
@click.option("-l", "--last", "last_opt", default="", help="last name")
@click.option("-p", "--position", "position_opt", default="", help="position")
@click.option("-c", "--company", "company_opt", default="", help="company")
@click.option("-s", "--server", "server_url", default="", help="custom server url (optional)")
@click.option("-k", "--api-key", "api_key", default="", help="custom api key (optional)")
@click.argument("firstName", required=False)
@click.argument("lastName", required=False)
@click.argument("position", required=False)
@click.argument("company", required=False)
async def identity_new(first_opt: str, last_opt: str, position_opt: str, company_opt: str, firstname: str, lastname: str, position: str, company: str, server_url: str, api_key: str):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot create an identity when impersonating a non admin key")
		return

	identity_details = datatypes.IdentityDetails()
	if (first_opt or firstname):
		identity_details.first_name = first_opt if first_opt else firstname
	if (last_opt or lastname):
		identity_details.last_name = last_opt if last_opt else lastname
	if (company_opt or company):
		identity_details.company = company_opt if company_opt else company
	if (position_opt or position):
		identity_details.position = position_opt if position_opt else position

	identity = await ClientSingleton.get_client().admin_identity_new(identity_details=identity_details, server_url=server_url, api_key=api_key)

	if ClientSingleton.is_script_mode_enabled():
		print(identity.id)
		return

	prompt: str = "identity creation"
	fg_color: str = "bright_blue"

	print_command_result(f"identity created with id: {identity.id}")

	try:
		# create associated client key
		client_key: datatypes.ClientKey = await ClientSingleton.get_client().admin_client_key_new(name=f"identity-new-key-{identity.id}", identity_id=identity.id)
		print_with_context(f"Here is your client key to connect to daemon with this identity:\n{click.style(client_key.key, underline=True)}", prompt=prompt, fg_color=fg_color)
		await asyncio.sleep(0.5)
		while not ask_question_with_context("Did you saved your client key ?", prompt=prompt, fg_color=fg_color):
			continue

		# ask to add this new identity as a contact
		if not ask_question_with_context("Do you want to add this identity to your contacts ?", prompt=prompt, fg_color=fg_color):
			ClientSingleton.set_current_identity_id(identity.id)
			click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")
			return

		# add this new identity as a contact
		discussion: datatypes.Discussion = await contact_new(identity_id=identity.id, prompt=prompt, fg_color=fg_color)

		ClientSingleton.set_current_identity_id(identity.id)
		click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")

		if discussion:
			print(f"You can now send messages to {discussion.title} in discussion {discussion.id}")
		else:
			print("Invitation process finished")
	except click.exceptions.Abort:
		pass


#####
# identity get
#####
# noinspection PyProtectedMember
@identity_tree.command("get", help="list identities on this daemon")
@click.option("-a", "--all", "get_all", is_flag=True)
# @click.option("-l", "--link", "show_invitation_link", is_flag=True)
@click.option("-f", "--fields", "fields", type=str)
@click.argument("identity_ids", nargs=-1, type=click.INT)
@click.option("--filter", "filter_", type=str)
async def identity_get(get_all: bool, identity_ids: tuple[int], fields: str, filter_: str = ""):
	if ClientSingleton.is_client_admin():
		# build filter
		identity_filter: datatypes.IdentityFilter = datatypes.IdentityFilter()
		if filter_:
			try:
				parsed_message = Parse(filter_, datatypes.IdentityFilter()._to_native(identity_filter))
				identity_filter = datatypes.IdentityFilter._from_native(parsed_message)
			except ParseError as e:
				print_error_message(f"Cannot parse filter: {e}")
				return

		if get_all or not identity_ids:
			async for identity in ClientSingleton.get_client().admin_identity_list(filter=identity_filter):
				# hide invitation link
				# if not show_invitation_link:
				# 	identity.invitation_url = ""
				filter_fields_and_print_normal_message(identity, fields)
		else:
			for identity_id in identity_ids:
				identity = await ClientSingleton.get_client().admin_identity_admin_get(identity_id=identity_id)
				# hide invitation link
				# if not show_invitation_link:
				# 	identity.invitation_url = ""
				filter_fields_and_print_normal_message(identity, fields)
	else:
		if identity_ids:
			print_error_message("Cannot list identities when impersonating a non admin key")
			return
		identity = await ClientSingleton.get_client().identity_get()
		# hide invitation link
		# if not show_invitation_link:
		# 	identity.invitation_url = ""
		filter_fields_and_print_normal_message(identity, fields)


#####
# identity rm
#####
@identity_tree.command("rm", help="delete identities")
@click.option("-a", "--all", "delete_all", is_flag=True)
@click.argument("identity_ids", nargs=-1, type=click.INT)
async def identity_delete(delete_all: bool, identity_ids: tuple[int]):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot delete an identity when impersonating a non admin key")
		return

	identity_ids = list(identity_ids)  # convert tuple to list
	if (not len(identity_ids) and not delete_all):
		raise click.exceptions.BadArgumentUsage("Specify identity id")

	if delete_all:
		identity_ids = [i.id async for i in ClientSingleton.get_client().admin_identity_list()]
		if not identity_ids:
			raise click.exceptions.BadArgumentUsage("No identity to delete")

	for identity_id in identity_ids:
		await ClientSingleton.get_client().admin_identity_delete(identity_id)
		print_command_result(f"Identity deleted: {identity_id}", identity_ids)


#####
# identity update
#####
@identity_tree.command("update", help="update your current identity details (name, position, ...)")
@click.option("-f", "--first", "first_opt", default="")
@click.option("-l", "--last", "last_opt", default="")
@click.option("-p", "--position", "position_opt", default="")
@click.option("-c", "--company", "company_opt", default="")
@click.argument("firstName", required=False)
@click.argument("lastName", required=False)
@click.argument("position", required=False)
@click.argument("company", required=False)
async def identity_update_details(first_opt: str, last_opt: str, position_opt: str, company_opt: str, firstname: str, lastname: str, position: str, company: str):
	identity_details = datatypes.IdentityDetails()
	if (first_opt or firstname):
		identity_details.first_name = first_opt if first_opt else firstname
	if (last_opt or lastname):
		identity_details.last_name = last_opt if last_opt else lastname
	if (company_opt or company):
		identity_details.company = company_opt if company_opt else company
	if (position_opt or position):
		identity_details.position = position_opt if position_opt else position

	if not identity_details.first_name and not identity_details.last_name:
		raise click.exceptions.BadArgumentUsage("Specify at least first or last name")

	await ClientSingleton.get_client().identity_update_details(identity_details)
	print_command_result("Identity updated")


#####
# identity key
#####
@identity_tree.command("key", help="set api key for current identity")
@click.argument("apiKey", nargs=1, required=True)
@click.option("-c", "--configuration", is_flag=True, required=False, help="Use a configuration link instead of a raw api key")
async def identity_apikey(configuration: bool, apikey: str):
	if configuration:
		result = await ClientSingleton.get_client().identity_set_configuration_link(configuration_link=apikey)
	else:
		result = await ClientSingleton.get_client().identity_set_api_key(api_key=apikey)
	print_command_result(f"Api key set up, new permissions: {result}")


#####
# identity photo
#####
@identity_tree.group("photo", help="manage current identity photo", cls=WrapperGroup)
def identity_photo_tree():
	pass


#####
# identity photo set
#####
@identity_photo_tree.command("set", help="set current identity photo")
@click.argument("photo_path", required=True, type=click.STRING)
async def identity_photo_set(photo_path):
	try:
		await ClientSingleton.get_client().identity_set_photo(photo_path)
		print_command_result("Identity photo set")
	except IOError as e:
		raise click.exceptions.BadArgumentUsage(str(e))


#####
# identity photo unset
#####
@identity_photo_tree.command("unset", help="remove current identity photo")
async def identity_photo_unset():
	await ClientSingleton.get_client().identity_remove_photo()
	print_command_result("Identity photo unset")


#####
# identity kc
#####
@identity_tree.group("kc", help="manage keycloak server for identities", cls=WrapperGroup)
def identity_kc_tree():
	pass


#####
# identity kc new
#####
@identity_kc_tree.command("new", help="create a new identity using a keycloak bot configuration link")
@click.argument("configuration_link", required=True, type=click.STRING)
async def identity_kc_new(configuration_link):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot create a new identity when impersonating a non admin key")
		return

	identity = await ClientSingleton.get_client().admin_identity_keycloak_new(configuration_link=configuration_link)

	if ClientSingleton.is_script_mode_enabled():
		print(identity.id)
		return

	prompt: str = "identity creation"
	fg_color: str = "bright_blue"

	print_command_result(f"identity created with id: {identity.id}")

	try:
		# create associated client key
		client_key: datatypes.ClientKey = await ClientSingleton.get_client().admin_client_key_new(name=f"identity-new-key-{identity.id}", identity_id=identity.id)
		print_with_context(f"Here is your client key to connect to daemon with this identity:\n{click.style(client_key.key, underline=True)}", prompt=prompt, fg_color=fg_color)
		await asyncio.sleep(0.5)
		while not ask_question_with_context("Did you saved your client key ?", prompt=prompt, fg_color=fg_color):
			continue

		ClientSingleton.set_current_identity_id(identity.id)
		click.secho(f"Now using identity: {ClientSingleton.get_current_identity_id()}", fg="green")
	except click.exceptions.Abort:
		pass



#####
# identity kc bind
#####
@identity_kc_tree.command("bind", help="attach your current identity to a keycloak server using a keycloak bot configuration link")
@click.argument("configuration_link", required=True, type=click.STRING)
async def identity_kc_bind(configuration_link):
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot manage identities when impersonating a non admin key")
		return

	await ClientSingleton.get_client().identity_keycloak_bind(configuration_link=configuration_link)
	print_command_result(f"Identity now linked to keycloak")


#####
# identity kc unbind
#####
@identity_kc_tree.command("unbind", help="remove keycloak server for current identity")
async def identity_kc_unbind():
	if not ClientSingleton.is_client_admin():
		print_error_message("Cannot manage identities when impersonating a non admin key")
		return

	await ClientSingleton.get_client().identity_keycloak_unbind()
	print_command_result(f"Identity is not linked to keycloak anymore")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/command/v1/discussion_commands.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/command/v1/discussion_commands.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import discussion_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_discussion__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1olvid/daemon/command/v1/discussion_commands.proto\x12\x17olvid.daemon.command.v1\x1a*olvid/daemon/datatypes/v1/discussion.proto\"l\n\x15\x44iscussionListRequest\x12H\n\x06\x66ilter\x18\x01 \x01(\x0b\x32+.olvid.daemon.datatypes.v1.DiscussionFilterH\x00R\x06\x66ilter\x88\x01\x01\x42\t\n\x07_filter\"a\n\x16\x44iscussionListResponse\x12G\n\x0b\x64iscussions\x18\x01 \x03(\x0b\x32%.olvid.daemon.datatypes.v1.DiscussionR\x0b\x64iscussions\";\n\x14\x44iscussionGetRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\"^\n\x15\x44iscussionGetResponse\x12\x45\n\ndiscussion\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.DiscussionR\ndiscussion\">\n\x1d\x44iscussionGetByContactRequest\x12\x1d\n\ncontact_id\x18\x01 \x01(\x04R\tcontactId\"g\n\x1e\x44iscussionGetByContactResponse\x12\x45\n\ndiscussion\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.DiscussionR\ndiscussion\"8\n\x1b\x44iscussionGetByGroupRequest\x12\x19\n\x08group_id\x18\x01 \x01(\x04R\x07groupId\"e\n\x1c\x44iscussionGetByGroupResponse\x12\x45\n\ndiscussion\x18\x01 \x01(\x0b\x32%.olvid.daemon.datatypes.v1.DiscussionR\ndiscussion\"\x85\x01\n\x16\x44iscussionEmptyRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\x12\x30\n\x11\x64\x65lete_everywhere\x18\x02 \x01(\x08H\x00R\x10\x64\x65leteEverywhere\x88\x01\x01\x42\x14\n\x12_delete_everywhere\"\x19\n\x17\x44iscussionEmptyResponse\"C\n\x1c\x44iscussionSettingsGetRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\"j\n\x1d\x44iscussionSettingsGetResponse\x12I\n\x08settings\x18\x01 \x01(\x0b\x32-.olvid.daemon.datatypes.v1.DiscussionSettingsR\x08settings\"i\n\x1c\x44iscussionSettingsSetRequest\x12I\n\x08settings\x18\x01 \x01(\x0b\x32-.olvid.daemon.datatypes.v1.DiscussionSettingsR\x08settings\"q\n\x1d\x44iscussionSettingsSetResponse\x12P\n\x0cnew_settings\x18\x01 \x01(\x0b\x32-.olvid.daemon.datatypes.v1.DiscussionSettingsR\x0bnewSettings\"\x1d\n\x1b\x44iscussionLockedListRequest\"g\n\x1c\x44iscussionLockedListResponse\x12G\n\x0b\x64iscussions\x18\x01 \x03(\x0b\x32%.olvid.daemon.datatypes.v1.DiscussionR\x0b\x64iscussions\"D\n\x1d\x44iscussionLockedDeleteRequest\x12#\n\rdiscussion_id\x18\x01 \x01(\x04R\x0c\x64iscussionId\" \n\x1e\x44iscussionLockedDeleteResponseB\xc6\x01\n\x1b\x63om.olvid.daemon.command.v1B\x17\x44iscussionCommandsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODC\xaa\x02\x17Olvid.Daemon.Command.V1\xca\x02\x17Olvid\\Daemon\\Command\\V1\xe2\x02#Olvid\\Daemon\\Command\\V1\\GPBMetadata\xea\x02\x1aOlvid::Daemon::Command::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.command.v1.discussion_commands_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.olvid.daemon.command.v1B\027DiscussionCommandsProtoP\001Z\017olvid.io/daemon\242\002\003ODC\252\002\027Olvid.Daemon.Command.V1\312\002\027Olvid\\Daemon\\Command\\V1\342\002#Olvid\\Daemon\\Command\\V1\\GPBMetadata\352\002\032Olvid::Daemon::Command::V1'
  _globals['_DISCUSSIONLISTREQUEST']._serialized_start=122
  _globals['_DISCUSSIONLISTREQUEST']._serialized_end=230
  _globals['_DISCUSSIONLISTRESPONSE']._serialized_start=232
  _globals['_DISCUSSIONLISTRESPONSE']._serialized_end=329
  _globals['_DISCUSSIONGETREQUEST']._serialized_start=331
  _globals['_DISCUSSIONGETREQUEST']._serialized_end=390
  _globals['_DISCUSSIONGETRESPONSE']._serialized_start=392
  _globals['_DISCUSSIONGETRESPONSE']._serialized_end=486
  _globals['_DISCUSSIONGETBYCONTACTREQUEST']._serialized_start=488
  _globals['_DISCUSSIONGETBYCONTACTREQUEST']._serialized_end=550
  _globals['_DISCUSSIONGETBYCONTACTRESPONSE']._serialized_start=552
  _globals['_DISCUSSIONGETBYCONTACTRESPONSE']._serialized_end=655
  _globals['_DISCUSSIONGETBYGROUPREQUEST']._serialized_start=657
  _globals['_DISCUSSIONGETBYGROUPREQUEST']._serialized_end=713
  _globals['_DISCUSSIONGETBYGROUPRESPONSE']._serialized_start=715
  _globals['_DISCUSSIONGETBYGROUPRESPONSE']._serialized_end=816
  _globals['_DISCUSSIONEMPTYREQUEST']._serialized_start=819
  _globals['_DISCUSSIONEMPTYREQUEST']._serialized_end=952
  _globals['_DISCUSSIONEMPTYRESPONSE']._serialized_start=954
  _globals['_DISCUSSIONEMPTYRESPONSE']._serialized_end=979
  _globals['_DISCUSSIONSETTINGSGETREQUEST']._serialized_start=981
  _globals['_DISCUSSIONSETTINGSGETREQUEST']._serialized_end=1048
  _globals['_DISCUSSIONSETTINGSGETRESPONSE']._serialized_start=1050
  _globals['_DISCUSSIONSETTINGSGETRESPONSE']._serialized_end=1156
  _globals['_DISCUSSIONSETTINGSSETREQUEST']._serialized_start=1158
  _globals['_DISCUSSIONSETTINGSSETREQUEST']._serialized_end=1263
  _globals['_DISCUSSIONSETTINGSSETRESPONSE']._serialized_start=1265
  _globals['_DISCUSSIONSETTINGSSETRESPONSE']._serialized_end=1378
  _globals['_DISCUSSIONLOCKEDLISTREQUEST']._serialized_start=1380
  _globals['_DISCUSSIONLOCKEDLISTREQUEST']._serialized_end=1409
  _globals['_DISCUSSIONLOCKEDLISTRESPONSE']._serialized_start=1411
  _globals['_DISCUSSIONLOCKEDLISTRESPONSE']._serialized_end=1514
  _globals['_DISCUSSIONLOCKEDDELETEREQUEST']._serialized_start=1516
  _globals['_DISCUSSIONLOCKEDDELETEREQUEST']._serialized_end=1584
  _globals['_DISCUSSIONLOCKEDDELETERESPONSE']._serialized_start=1586
  _globals['_DISCUSSIONLOCKEDDELETERESPONSE']._serialized_end=1618
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/datatypes/v1/client_key.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/datatypes/v1/client_key.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*olvid/daemon/datatypes/v1/client_key.proto\x12\x19olvid.daemon.datatypes.v1\"R\n\tClientKey\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x10\n\x03key\x18\x02 \x01(\tR\x03key\x12\x1f\n\x0bidentity_id\x18\x03 \x01(\x04R\nidentityId\"\xb4\x01\n\x0f\x43lientKeyFilter\x12\x1d\n\tadmin_key\x18\x01 \x01(\x08H\x00R\x08\x61\x64minKey\x12!\n\x0bidentity_id\x18\x02 \x01(\x04H\x00R\nidentityId\x12$\n\x0bname_search\x18\x03 \x01(\tH\x01R\nnameSearch\x88\x01\x01\x12\x15\n\x03key\x18\x04 \x01(\tH\x02R\x03key\x88\x01\x01\x42\n\n\x08identityB\x0e\n\x0c_name_searchB\x06\n\x04_keyB\xc7\x01\n\x1d\x63om.olvid.daemon.datatypes.v1B\x0e\x43lientKeyProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODD\xaa\x02\x19Olvid.Daemon.Datatypes.V1\xca\x02\x19Olvid\\Daemon\\Datatypes\\V1\xe2\x02%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\xea\x02\x1cOlvid::Daemon::Datatypes::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.datatypes.v1.client_key_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\035com.olvid.daemon.datatypes.v1B\016ClientKeyProtoP\001Z\017olvid.io/daemon\242\002\003ODD\252\002\031Olvid.Daemon.Datatypes.V1\312\002\031Olvid\\Daemon\\Datatypes\\V1\342\002%Olvid\\Daemon\\Datatypes\\V1\\GPBMetadata\352\002\034Olvid::Daemon::Datatypes::V1'
  _globals['_CLIENTKEY']._serialized_start=73
  _globals['_CLIENTKEY']._serialized_end=155
  _globals['_CLIENTKEYFILTER']._serialized_start=158
  _globals['_CLIENTKEYFILTER']._serialized_end=338
# @@protoc_insertion_point(module_scope)



import os
from google.protobuf.json_format import Parse, ParseError

from ..interactive_tree import interactive_tree
from ..tools.cli_tools import *
from ..tools.click_wrappers import WrapperGroup


#####
# attachment
#####
@interactive_tree.group("attachment", help="list and save your attachments", cls=WrapperGroup)
def attachment_tree():
	pass


#####
# attachment get
#####
# noinspection PyProtectedMember
@attachment_tree.command("get", help="list current identity attachments")
@click.option("-a", "--all", "get_all", is_flag=True, help="get attachments for current owned identity")
@click.option("-m", "--message", "message_id", type=click.STRING, help="get attachments associated to a message id")
@click.argument("attachment_ids", nargs=-1, required=False, type=click.STRING)
@click.option("--filter", "filter_", type=str)
@click.option("-f", "--fields", "fields", type=str)
async def attachment_get(get_all: bool, message_id: str, attachment_ids: tuple[str], fields: str, filter_: str = ""):
	# build filter
	attachment_filter: datatypes.AttachmentFilter = datatypes.AttachmentFilter()
	if filter_:
		try:
			parsed_attachment = Parse(filter_, datatypes.AttachmentFilter()._to_native(attachment_filter))
			attachment_filter = datatypes.AttachmentFilter._from_native(parsed_attachment)
		except ParseError as e:
			print_error_message(f"Cannot parse filter: {e}")
			return

	if message_id:
		async for attachment in ClientSingleton.get_client().attachment_message_list(message_id=string_to_message_id(message_id)):
			filter_fields_and_print_normal_message(attachment, fields)
	elif get_all or not attachment_ids:
		async for attachment in ClientSingleton.get_client().attachment_list(filter=attachment_filter):
			filter_fields_and_print_normal_message(attachment, fields)
	else:
		for attachment_id in attachment_ids:
			attachment = await ClientSingleton.get_client().attachment_get(attachment_id=string_to_attachment_id(attachment_id))
			filter_fields_and_print_normal_message(attachment, fields)


#####
# attachment rm
#####
@attachment_tree.command("rm", help="delete attachments")
@click.argument("attachment_ids", nargs=-1, type=click.STRING, required=True)
async def attachment_rm(attachment_ids: tuple[str]):
	for attachment_id in attachment_ids:
		await ClientSingleton.get_client().attachment_delete(string_to_attachment_id(attachment_id))
		print_command_result(f"attachment deleted: {attachment_id}")


#####
# attachment save
#####
@attachment_tree.command("save", help="download an attachment file and save it in your filesystem")
@click.argument("attachment_id", nargs=1, type=click.STRING)
@click.option("-p", "--path", "path", help="directory to store downloaded attachment", nargs=1, type=click.STRING, required=False)
@click.option("-f", "--filename", "filename", help="override original file name", nargs=1, type=click.STRING, required=False)
async def attachment_save(attachment_id: str, path: str, filename: str):
	# ues the default path to store attachments (and create dir if needed)
	if not path:
		path = "./attachments"
		if not os.path.isdir(path):
			os.mkdir(path)
	# check the path exists
	elif not os.path.isdir:
		raise click.exceptions.FileError("{}: no such directory".format(path))

	if not filename:
		attachment = await ClientSingleton.get_client().attachment_get(attachment_id=string_to_attachment_id(attachment_id))
		filename = attachment.file_name
	try:
		# create file
		with open(os.path.join(path, filename), "wb") as fd:
			async for chunk in ClientSingleton.get_client().attachment_download(attachment_id=string_to_attachment_id(attachment_id)):
				fd.write(chunk)
		print_success_message(f"Attachment downloaded at: {os.path.join(path, filename)}", f"{os.path.join(path, filename)}")
	except Exception as e:
		raise click.exceptions.BadArgumentUsage("Unable to download file: " + str(e))



import re
import logging

from olvid import datatypes, listeners, OlvidBot

from ClientHolder import ClientHolder, ClientWrapper
from utils import send_message_wait_and_check_content, send_message_with_attachments_in_memory_and_check


def compare_message_lists(given: list[datatypes.Message], expected: list[datatypes.Message]):
	given = sorted(given, key=message_sort_key)
	expected = sorted(expected, key=message_sort_key)
	assert len(given) == len(expected), f"Invalid list length: {len(given)} != {len(expected)}\ngiven:{given}\nexpected: {expected}"
	for i in range(len(given)):
		assert given[i] == expected[i], f"Invalid element in list: {given[i]} != {expected[i]}\ngiven:{given}\nexpected: {expected}"


def message_sort_key(m: datatypes.Message):
	return m.id.id if m.id.type == datatypes.MessageId.Type.TYPE_INBOUND else -1 * m.id.id


async def test_filtering_discussion(c1: ClientWrapper, c2: ClientWrapper):
	# create discussions
	created_groups: list[datatypes.Group] = [
		await c1.group_new_standard_group(name="GROUP-NAME"),
		await c1.group_new_standard_group(name="MyNameIs1"),
		await c1.group_new_standard_group(name="MyNameIs2"),
	]

	all_discussions: list[datatypes.Discussion] = sorted([d async for d in c1.discussion_list()], key=lambda d: d.id)

	# delete created discussions
	for group in created_groups:
		await c1.group_disband(group_id=group.id)


async def test_filtering_invitation(c1: ClientWrapper, c2: ClientWrapper):
	invitations: list[datatypes.Invitation] = [i async for i in c1.invitation_list(datatypes.InvitationFilter(type=datatypes.InvitationFilter.Type.TYPE_INTRODUCTION))]


async def test_filtering_messages(c1: ClientWrapper, c2: ClientWrapper):
	contact_1, contact_2 = await c1.get_contact_associated_to_another_client(c2), await c1.get_contact_associated_to_another_client(c2)

	# create messages
	await send_message_wait_and_check_content(c1, c2, "Hello there")
	await send_message_wait_and_check_content(c1, c2, "my body is 1")
	message_to_react_to_1, _ = await send_message_wait_and_check_content(c1, c2, "Please react to me")
	_, message_to_react_to_2 = await send_message_wait_and_check_content(c1, c2, "I want reactions too")
	message_to_reply_to, _ = await send_message_wait_and_check_content(c1, c2, "MY body is 2")
	await send_message_wait_and_check_content(c1, c2, "This message embed a reply", replied_message_id=message_to_reply_to.id)
	await send_message_with_attachments_in_memory_and_check(c1, c2, body="", filenames=["file.txt", "image.jpg"], file_contents=[b"a", b"d"])
	await send_message_with_attachments_in_memory_and_check(c1, c2, body="Film and image", filenames=["film.mp4", "image.jpg"], file_contents=[b"a", b"d"])

	# send reactions
	# add bot for reactions
	bots: list[OlvidBot] = [
		c1.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=2)),
		c2.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=2)),

		# TODO re-enable when notification filtering implemented
		# c1.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=1, filter=datatypes.MessageFilter(reacted_by_me=True, reaction="🌍"))),
		# c2.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=1, filter=datatypes.MessageFilter(reacted_by_contact_id=contact_2.id, reaction="🌍"))),
		# c1.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=1, filter=datatypes.MessageFilter(reacted_by_contact_id=contact_1.id, reaction="🪦"))),
		# c2.create_notification_bot(listeners.MessageReactionAddedListener(handler=lambda m, r: m, count=1, filter=datatypes.MessageFilter(reacted_by_me=True, reaction="🪦"))),
	]
	await message_to_react_to_1.react("🌍")
	await message_to_react_to_2.react("🪦")

	for bot in bots:
		await bot.wait_for_listeners_end()

	####
	# start tests
	####
	all_messages_1: list[datatypes.Message] = sorted([m async for m in c1.message_list()], key=message_sort_key)
	all_messages_2: list[datatypes.Message] = sorted([m async for m in c2.message_list()], key=message_sort_key)

	# no filter
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter()
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1])

	# body search
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(body_search="body")
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if re.findall(r"body", m.body)])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(body_search="nothing you can find in a body")
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if re.findall(r"nothing you can find in a body", m.body)])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(body_search="^my")
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if re.findall(r"^my", m.body)])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(body_search="")
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if re.findall(r"", m.body)])

	# location
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(location=datatypes.MessageFilter.Location.LOCATION_HAVE)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if m.message_location])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(location=datatypes.MessageFilter.Location.LOCATION_HAVE_NOT)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if not m.message_location])

	# reply
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(do_not_reply_to_a_message=True)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if not m.replied_message_id])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(reply_to_a_message=True)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if m.replied_message_id])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(replied_message_id=message_to_reply_to.id)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if m.replied_message_id == message_to_reply_to.id])
	message_filter: datatypes.MessageFilter = datatypes.MessageFilter(replied_message_id=message_to_react_to_1.id)
	compare_message_lists([m async for m in c1.message_list(filter=message_filter)], [m for m in all_messages_1 if m.replied_message_id == message_to_react_to_1.id])

	# reactions
	# noinspection PyShadowingBuiltins
	async def test_reaction(client: ClientWrapper, other_client_contact_id: int, all_messages: list[datatypes.Message]):
		# reaction
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reaction=datatypes.MessageFilter.Reaction.REACTION_HAS)
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reaction=datatypes.MessageFilter.Reaction.REACTION_HAS_NOT)
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions]) == 0])

		# reaction filters
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="🌍")])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "🌍"]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="🌍", reacted_by_me=True)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "🌍" and r.contact_id == 0]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="🪦")])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "🪦"]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="🪦", reacted_by_contact_id=other_client_contact_id)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "🪦" and r.contact_id == other_client_contact_id]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="💣")])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "💣"]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="💣", reacted_by_me=True)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "💣" and r.contact_id == 0]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reaction="💣", reacted_by_contact_id=other_client_contact_id)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.reaction == "💣" and r.contact_id == other_client_contact_id]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reacted_by_me=True)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.contact_id == 0]) > 0])
		filter: datatypes.MessageFilter = datatypes.MessageFilter(reactions_filter=[datatypes.ReactionFilter(reacted_by_contact_id=other_client_contact_id)])
		compare_message_lists([m async for m in client.message_list(filter=filter)], [m for m in all_messages if len([r for r in m.reactions if r.contact_id == other_client_contact_id]) > 0])

	await test_reaction(c1, contact_1.id, all_messages_1)
	await test_reaction(c2, contact_2.id, all_messages_2)


async def test_filtering(client_holder: ClientHolder, fast_mode=False):
	for c1, c2 in client_holder.get_all_client_pairs():
		logging.info(f"{c1.identity.id} <-> {c2.identity.id}: message filtering")
		await test_filtering_messages(c1, c2)



import logging
from grpc.aio import AioRpcError

from ClientHolder import ClientHolder, ClientWrapper
from olvid import OlvidClient, OlvidBot
from olvid import datatypes, listeners


# a bot that accept next group invitation it will receive
class GroupAcceptBot(OlvidBot):
	def __init__(self, identity_id: int, parent_client: OlvidClient):
		super().__init__(f"GroupAcceptBot-{identity_id}", parent_client=parent_client)
		self._identity_id = identity_id

		# add a listener for group invitations only
		# noinspection PyTypeChecker
		self.add_listener(listeners.InvitationReceivedListener(handler=self.invitation_handler, checkers=[lambda i: i.status == datatypes.Invitation.Status.STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT], count=1))

	async def invitation_handler(self, invitation: datatypes.Invitation):
		if invitation.status == invitation.status.STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT:
			await self.invitation_accept(invitation_id=invitation.id)
		else:
			logging.error(f"{self._identity_id}: received an invalid invitation type: {invitation.status}", invitation)


# a bot that decline next group invitation it will receive
class GroupDeclineBot(GroupAcceptBot):
	async def invitation_handler(self, invitation: datatypes.Invitation):
		if invitation.status == invitation.status.STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT:
			await self.invitation_decline(invitation_id=invitation.id)
		else:
			logging.error(f"{self._identity_id}: received an invalid invitation type: {invitation.status}", invitation)


# noinspection PyProtectedMember
def get_group_seen_by_a_member(original_group: datatypes.Group, member_contact_id: int):
	group: datatypes.Group = original_group._clone()

	# exchange member and owner permissions
	contact_member: datatypes.GroupMember = [m for m in group.members if m.contact_id == member_contact_id][0]
	contact_permissions: datatypes.GroupMemberPermissions = contact_member.permissions
	contact_member.permissions = group.own_permissions
	group.own_permissions = contact_permissions

	# hide ids
	group.id = 0
	for member in group.members:
		member.contact_id = 0
	for pending_member in group.pending_members:
		pending_member.contact_id = 0
		pending_member.pending_member_id = 0

	return group


async def test_create_normal_group(client_holder: ClientHolder, admin_client: ClientWrapper, group_name: str, group_description: str) -> datatypes.Group:
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in member_clients]

	# create notif bots for admin
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(listeners.GroupNewListener(handler=admin_client.get_check_content_handler(datatypes.Group(name=group_name, description=group_description), notification_type=listeners.NOTIFICATIONS.GROUP_NEW), count=1)),
		admin_client.create_notification_bot(listeners.DiscussionNewListener(handler=admin_client.get_check_content_handler(datatypes.Discussion(title=group_name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_NEW), count=1)),
		admin_client.create_notification_bot(listeners.GroupUpdateInProgressListener(handler=lambda gid: gid, count=1)),
		admin_client.create_notification_bot(listeners.GroupUpdateFinishedListener(handler=lambda gid: gid, count=1)),
	]
	# check content of group member joined notif
	for member_client in member_clients:
		member_contact: datatypes.Contact = member_contacts[member_clients.index(member_client)]

		def get_contact_id_checker(contact_id: int):
			# noinspection PyUnusedLocal
			def checker(a, b):
				return b.contact_id == contact_id
			return checker

		bots.extend([
			admin_client.create_notification_bot(listeners.GroupMemberJoinedListener(handler=admin_client.get_check_content_handler(datatypes.Group(name=group_name, description=group_description), datatypes.GroupMember(contact_id=member_contact.id), notification_type=listeners.NOTIFICATIONS.GROUP_MEMBER_JOINED), checkers=[get_contact_id_checker(member_contact.id)], count=1)),
			admin_client.create_notification_bot(listeners.GroupPendingMemberRemovedListener(handler=admin_client.get_check_content_handler(datatypes.Group(name=group_name, description=group_description), datatypes.PendingGroupMember(contact_id=member_contact.id, display_name=member_contact.display_name), notification_type=listeners.NOTIFICATIONS.GROUP_PENDING_MEMBER_REMOVED), checkers=[get_contact_id_checker(member_contact.id)], count=1)),
		])

	# create bots to accept invitations and check notif on member side
	for member_client in member_clients:
		bots.extend([
			# bot to accept invitation
			GroupAcceptBot(identity_id=member_client.identity.id, parent_client=member_client),
			# check content bots
			# invitations
			member_client.create_notification_bot(listeners.InvitationReceivedListener(handler=admin_client.get_check_content_handler(datatypes.Invitation(display_name=group_name, status=datatypes.Invitation.Status.STATUS_GROUP_INVITATION_WAIT_YOU_TO_ACCEPT, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_RECEIVED), count=1)),
			member_client.create_notification_bot(listeners.InvitationDeletedListener(handler=admin_client.get_check_content_handler(datatypes.Invitation(display_name=group_name, status=datatypes.Invitation.Status.STATUS_UNSPECIFIED, sas=""), notification_type=listeners.NOTIFICATIONS.INVITATION_DELETED), count=1)),
			# group
			member_client.create_notification_bot(listeners.GroupNewListener(handler=admin_client.get_check_content_handler(datatypes.Group(name=group_name, description=group_description), notification_type=listeners.NOTIFICATIONS.GROUP_NEW), count=1)),
			member_client.create_notification_bot(listeners.DiscussionNewListener(handler=admin_client.get_check_content_handler(datatypes.Discussion(title=group_name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_NEW), count=1)),
			# group members (only check count)
			member_client.create_notification_bot(listeners.GroupMemberJoinedListener(handler=lambda a, b: a, count=len(member_clients))),
			member_client.create_notification_bot(listeners.GroupPendingMemberRemovedListener(handler=lambda a, b: a, count=len(member_clients))),
		])

	# create group
	new_group: datatypes.Group = await admin_client.group_new_controlled_group(name=group_name, description=group_description, contact_ids=[c.id for c in member_contacts])

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()

	new_group._test_assertion(datatypes.Group(name=group_name, description=group_description, keycloak_managed=False, has_a_photo=False))
	assert len(new_group.pending_members) == len(member_clients), f"there are not enough pending members in group ({len(member_clients)}) {new_group}"

	return new_group


async def test_disband_group(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in
												member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	# create notif-checker bots for group admin
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(listeners.GroupDeletedListener(handler=admin_client.get_check_content_handler(original_group, notification_type=listeners.NOTIFICATIONS.GROUP_DELETED), count=1)),
		admin_client.create_notification_bot(listeners.DiscussionLockedListener(handler=admin_client.get_check_content_handler(datatypes.Discussion(title=original_group.name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
		admin_client.create_notification_bot(listeners.GroupUpdateInProgressListener(handler=lambda gid: gid, count=1)),
		admin_client.create_notification_bot(listeners.GroupUpdateFinishedListener(handler=lambda gid: gid, count=1)),
	]

	# TODO remove this, used to pass a known bug (group type is not stored so we cannot retrieve it when group have been disbanded)
	original_group.type = datatypes.Group.Type.TYPE_UNSPECIFIED

	# create notif-checker bots for members
	for i in range(len(member_clients)):
		member_client = member_clients[i]
		member_contact_id = member_contacts[i].id
		bots.extend([
			member_client.create_notification_bot(listeners.GroupDeletedListener(handler=member_client.get_check_content_handler(get_group_seen_by_a_member(original_group, member_contact_id), notification_type=listeners.NOTIFICATIONS.GROUP_DELETED), count=1)),
			member_client.create_notification_bot(listeners.DiscussionLockedListener(handler=member_client.get_check_content_handler(datatypes.Discussion(title=original_group.name), notification_type=listeners.NOTIFICATIONS.DISCUSSION_LOCKED), count=1)),
		])

	disbanded_group = await admin_client.group_disband(group_id=original_group.id)
	disbanded_group._test_assertion(original_group)

	for bot in bots:
		await bot.wait_for_listeners_end()
		await bot.stop()


async def test_try_do_delete_a_contact_in_group(client_holder: ClientHolder, admin_client: ClientWrapper):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in member_clients]

	for member_contact in member_contacts:
		try:
			await admin_client.contact_delete(contact_id=member_contact.id)
			assert False, "Was able to delete a contact in a group"
		except AioRpcError:
			continue


async def test_concurrent_group_updates(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	group_updated_1: datatypes.Group = original_group._clone()
	group_updated_1.name += "-First Update"
	group_updated_1.description += "-First Update"
	group_updated_2: datatypes.Group = original_group._clone()
	group_updated_2.name = ""
	group_updated_2.description = ""

	# this handler is used to check multiple ordered notifications with different content
	def get_recursive_check_content_handler(client: ClientWrapper, listener_type: type[listeners.GenericNotificationListener], expectations: list[tuple], notification_type: listeners.NOTIFICATIONS):
		def recursive_check_content_handler(*messages):
			if not expectations:
				assert False, "Invalid expectations count"
			handler = client.get_check_content_handler(*(expectations[0]), notification_type=notification_type)
			ret = handler(*messages)

			# create next listener
			if len(expectations) > 1:
				# noinspection PyArgumentList
				client.create_notification_bot(listener_type(handler=get_recursive_check_content_handler(client, listener_type=listener_type, expectations=expectations[1:], notification_type=notification_type), count=1))

			return ret
		return recursive_check_content_handler

	# wait for task 1 and associated notif
	# create notif-checker bots for group admin
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupNameUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.GroupNameUpdatedListener, [(group_updated_1, original_group.name), (group_updated_2, group_updated_1.name)], notification_type=listeners.NOTIFICATIONS.GROUP_NAME_UPDATED),
				count=1)),
		admin_client.create_notification_bot(
			listeners.GroupDescriptionUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.GroupDescriptionUpdatedListener, [(group_updated_1, original_group.description), (group_updated_2, group_updated_1.description)], notification_type=listeners.NOTIFICATIONS.GROUP_DESCRIPTION_UPDATED),
				count=1)),
		admin_client.create_notification_bot(
			listeners.DiscussionTitleUpdatedListener(
				handler=get_recursive_check_content_handler(admin_client, listeners.DiscussionTitleUpdatedListener, [(datatypes.Discussion(title=group_updated_1.name), original_group.name), (datatypes.Discussion(title=group_updated_2.name), group_updated_1.name)], notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED),
				count=1)),
	]

	# create notif-checker bots for members: members only receive one update notifications (updates are probably too fast to give time to handle intermediary blob)
	for i in range(len(member_clients)):
		member_client = member_clients[i]
		member_contact_id = member_contacts[i].id
		bots.extend([
			member_client.create_notification_bot(
				listeners.GroupNameUpdatedListener(
					handler=member_client.get_check_content_handler(get_group_seen_by_a_member(group_updated_2, member_contact_id), original_group.name, notification_type=listeners.NOTIFICATIONS.GROUP_NAME_UPDATED),
					count=1)),
			member_client.create_notification_bot(
				listeners.GroupDescriptionUpdatedListener(
					handler=member_client.get_check_content_handler(get_group_seen_by_a_member(group_updated_2, member_contact_id), original_group.description, notification_type=listeners.NOTIFICATIONS.GROUP_DESCRIPTION_UPDATED),
					count=1)),
			member_client.create_notification_bot(
				listeners.DiscussionTitleUpdatedListener(
					handler=member_client.get_check_content_handler(datatypes.Discussion(title=group_updated_2.name), original_group.name, notification_type=listeners.NOTIFICATIONS.DISCUSSION_TITLE_UPDATED),
					count=1)),
		])

	# launch concurrent tasks
	task1 = admin_client.group_update(group=group_updated_1)
	task2 = admin_client.group_update(group=group_updated_2)

	group_update_response_1: datatypes.Group = await task1
	group_update_response_1._test_assertion(group_updated_1)

	group_update_response_2: datatypes.Group = await task2
	group_update_response_2._test_assertion(group_updated_2)

	for bot in bots:
		await bot.wait_for_listeners_end()


# noinspection DuplicatedCode
async def test_update_member_permissions(client_holder: ClientHolder, admin_client: ClientWrapper, group_id: int):
	member_clients: list[ClientWrapper] = [c for c in client_holder.clients if c != admin_client]
	member_contacts: list[datatypes.Contact] = [await admin_client.get_contact_associated_to_another_client(c) for c in
												member_clients]

	original_group: datatypes.Group = await admin_client.group_get(group_id=group_id)

	if not original_group.own_permissions.admin:
		return

	# choose a member, save it's permissions and update it
	group: datatypes.Group = original_group._clone()
	member_contact_id: int = member_contacts[0].id
	member_index: int = [i for i in range(len(group.members)) if group.members[i].contact_id == member_contact_id][0]
	original_member_permissions: datatypes.GroupMemberPermissions = group.members[member_index].permissions._clone()
	new_member_permissions: datatypes.GroupMemberPermissions = datatypes.GroupMemberPermissions()
	group.members[member_index].permissions = new_member_permissions

	# prepare notification handlers
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupMemberPermissionsUpdatedListener(
				handler=admin_client.get_check_content_handler(group, group.members[member_index], original_member_permissions, notification_type=listeners.NOTIFICATIONS.GROUP_MEMBER_PERMISSIONS_UPDATED),
				count=1)),
	]

	# update member permissions
	response_group = await admin_client.group_update(group=group)
	response_group._test_assertion(group)

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()

	####
	# restore original permissions
	####
	group.members[member_index].permissions = original_member_permissions
	bots: list[OlvidClient] = [
		admin_client.create_notification_bot(
			listeners.GroupMemberPermissionsUpdatedListener(
				handler=admin_client.get_check_content_handler(group, group.members[member_index], new_member_permissions, notification_type=listeners.NOTIFICATIONS.GROUP_MEMBER_PERMISSIONS_UPDATED),
				count=1)),
	]

	# update member permissions
	response_group = await admin_client.group_update(group=group)
	response_group._test_assertion(group)

	# wait for notifications
	for bot in bots:
		await bot.wait_for_listeners_end()


async def test_group(client_holder: ClientHolder):
	for client in client_holder.clients:
		logging.info(f"{client.identity.id}: create normal group")
		group = await test_create_normal_group(client_holder, client, f"{client.identity.id}-NormalGroupName", group_description=f"{client.identity.id} normal group description")
		group_id: int = group.id  # do not use original group, it is not up to date

		logging.info(f"{client.identity.id}: delete contact in a group")
		await test_try_do_delete_a_contact_in_group(client_holder, client)
		logging.info(f"{client.identity.id}: concurrent group updates")
		await test_concurrent_group_updates(client_holder, client, group_id)
		logging.info(f"{client.identity.id}: permission updates")
		await test_update_member_permissions(client_holder, client, group_id)

		logging.info(f"{client.identity.id}: disband normal group")
		await test_disband_group(client_holder, client, group_id)

		logging.info(f"{client.identity.id}: create empty name group")
		group = await test_create_normal_group(client_holder, client, "", "")
		group_id: int = group.id  # do not use original group, it is not up to date
		logging.info(f"{client.identity.id}: disband empty name group")
		await test_disband_group(client_holder, client, group_id)
